package edu.blinded.authors.tdrExperimentAnalysis;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class SessionData {

    public static interface Defect {

        public static<T extends Defect> void putDef(
                        Map<String, List<Defect>> ret, String remark, T defectClassification, T... furtherDefects) {
            final List<Defect> defects = new ArrayList<>();
            defects.add(defectClassification);
            defects.addAll(Arrays.asList(furtherDefects));
            ret.put(remark, defects);
        }

        public abstract String getShortString();

        public abstract boolean isDelocalized();

    }

    public enum FP implements Defect {
        FP;

        @Override
        public String getShortString() {
            return this.getClass().getSimpleName();
        }

        @Override
        public boolean isDelocalized() {
            return false;
        }
    }

    private static final Pattern WMSPAN_TRIAL_PATTERN = Pattern.compile("trial result,([0-9]+),([0-9]+),([0-9]+)");

    private final Set<String> DUPLICATE_BLACKLIST = new HashSet<>(Arrays.asList(
                    "239824586668", "239824586671",
                    "239373198651", "239373198683"));



    private final List<LogRecord> records;
    private final Map<Integer, List<LogRecord>> reviewRecordCache = new HashMap<>();

    public SessionData(List<LogRecord> records) {
        this.records = records;
        Collections.sort(this.records, new Comparator<LogRecord>() {
            @Override
            public int compare(LogRecord o1, LogRecord o2) {
                final int cmp = Long.compare(o1.getTimestamp(), o2.getTimestamp());
                if (cmp != 0) {
                    return cmp;
                }
                if (o1.isFromReview() && o2.isFromReview()) {
                    return Long.compare(o1.getClientTimeReview(), o2.getClientTimeReview());
                } else {
                    return 0;
                }
            }
        });
    }

    public String getId() {
        return this.records.get(0).getSessionId();
    }

    public String getShortId() {
        return this.getRestAfterMarker("alpha id is ", null);
    }

    private String getRestAfterMarker(final String marker, String defaultValue) {
        final Optional<LogRecord> value = this.findFirstStartingWith(marker);
        if (!value.isPresent() && defaultValue != null) {
            return defaultValue;
        }
        return value.get().getRest().substring(marker.length());
    }

    private Optional<LogRecord> findFirstStartingWith(String prefix) {
        return this.records.stream().filter((LogRecord l) -> l.getRest().startsWith(prefix)).findFirst();
    }

    public long getTotalDuration() {
        LogRecord lastNonCancelRecord = null;
        for (int i = this.records.size() - 1; i > 0; i--) {
            final LogRecord cur = this.records.get(i);
            final String rest = cur.getRest().trim();
            if (!(rest.equals("experiment cancelled") || rest.equals("change state;cancelled"))) {
                lastNonCancelRecord = cur;
                break;
            }
        }
        return this.findFirstStartingWith("experiment started").get().timeUntil(lastNonCancelRecord);
    }

    public String getVariable(String name) {
        return this.findFirstStartingWith("set variable;" + name + ";").get().getSuffixOfRest(2);
    }

    public String getVariable(String name, String defaultValue) {
        final Optional<LogRecord> record = this.findFirstStartingWith("set variable;" + name + ";");
        return record.isPresent() ? record.get().getSuffixOfRest(2) : defaultValue;
    }

    public int getUnderstandingCorrect(int reviewNumber) {
        switch (reviewNumber) {
        case 0:
        case 1:
            return this.getChange(reviewNumber) == PatchToDiff.P1 ? this.getUnderstandingCorrectC1() : this.getUnderstandingCorrectC2();
        default:
            throw new AssertionError();
        }
    }

    public int getUnderstandingTotal(int reviewNumber) {
//        switch (reviewNumber) {
//        case 0:
//        case 1:
//            return this.getChange(reviewNumber) == PatchToDiff.P1 ? 4 : 3;
//        default:
//            throw new AssertionError();
//        }
        return 3;
    }

    public int getUnderstandingCorrectWarmup() {
        return this.questionCorrectWA()
            + this.questionCorrectWB();
    }

    public int questionCorrectWB() {
        return toInt(this.getVariable("methodAdded", "").equals("1"));
    }

    public int questionCorrectWA() {
        return toInt(this.getVariable("idStatement", "").equals("2"));
    }

    public int getUnderstandingCorrectC1() {
        return this.questionCorrectC1A()
            + this.questionCorrectC1B()
            + this.questionCorrectC1C();
    }

    public int questionCorrectC1A() {
        return toInt(this.getVariable("playerIntroduced", "").equals("1"));
    }

    public int questionCorrectC1B() {
        return toInt(this.getVariable("playerNull", "").equals("0"));
    }

    public int questionCorrectC1C() {
        return toInt(this.getVariable("setupInstructions", "").equals("1"));
    }

    public int getUnderstandingCorrectC2() {
        return this.questionCorrectC2A()
            + this.questionCorrectC2B()
            + this.questionCorrectC2C();
    }

    public int questionCorrectC2A() {
        return toInt(this.getVariable("refactoring", "").equals("0"));
    }

    public int questionCorrectC2B() {
        return toInt(this.getVariable("imageSprite", "").equals("0"));
    }

    public int questionCorrectC2C() {
        return toInt(this.getVariable("behaviourSplit", "").equals("0"));
    }

    private static int toInt(boolean bool) {
        return bool ? 1 : 0;
    }

    private static final class ReviewStateTracer {
        private final boolean includeIdling;
        private final boolean includeTests;
        private boolean reviewActive = true;
        private boolean pauseActive = false;
        private boolean descActive = false;
        private boolean idlingActive = false;
        private boolean testActive;
        private long activeTime;
        private LogRecord prev;

        public ReviewStateTracer(LogRecord firstRecord, boolean includeIdling) {
            this(firstRecord, includeIdling, true, false);
        }

        public ReviewStateTracer(LogRecord firstRecord, boolean includeIdling, boolean includeTests, boolean firstIsTest) {
            this.prev = firstRecord;
            this.includeIdling = includeIdling;
            this.includeTests = includeTests;
            this.testActive = firstIsTest;
        }

        public void handle(LogRecord r) {
            final boolean wasActive = this.isActive();

            final String rest = r.getSuffixOfRest(3);
            if (rest.startsWith("idling started")) {
                this.idlingActive = true;
            } else if (rest.startsWith("idling ended")) {
                this.idlingActive = false;
            } else if (rest.startsWith("pause started")) {
                this.pauseActive = true;
            } else if (rest.startsWith("pause ended")) {
                this.pauseActive = false;
            } else if (rest.startsWith("desc started")) {
                this.descActive = true;
            } else if (rest.startsWith("desc ended")) {
                this.descActive = false;
            } else if (rest.startsWith("review ended")) {
                this.reviewActive = false;
            } else if (rest.startsWith("pressedToggleShownCode")) {
                this.testActive = !this.testActive;
            }

            final boolean willBeActive = this.isActive();

            if (wasActive != willBeActive) {
                if (wasActive) {
                    this.activeTime += this.prev.reviewTimeUntil(r);
                    this.prev = null;
                } else {
                    this.prev = r;
                }
            }
        }

        private boolean isActive() {
            final boolean idling;
            if (this.includeIdling) {
                //include idling := include idle time in sum => don't respect idle in active check
                idling = false;
            } else {
                idling = this.idlingActive;
            }
            final boolean inTests;
            if (this.includeTests) {
                //include tests := include time spent reviewing tests in sum => don't respect testActive in active check
                inTests = false;
            } else {
                inTests = this.testActive;
            }
            return this.reviewActive && !this.pauseActive && !this.descActive && !idling && !inTests;
        }

        public long timeUntil(LogRecord r) {
            if (this.isActive()) {
                return this.activeTime + this.prev.reviewTimeUntil(r);
            } else {
                return this.activeTime;
            }
        }

    }

    public long getNetTimeForReview(int number, boolean includeIdling) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        if (reviewRecords.isEmpty()) {
            return -1;
        }
        final ReviewStateTracer state = new ReviewStateTracer(reviewRecords.get(0), includeIdling);
        for (final LogRecord r : reviewRecords) {
            state.handle(r);
        }
        return state.activeTime;
    }

    public long getProdTimeForReview(int number) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        if (reviewRecords.isEmpty()) {
            return -1;
        }
        final Treatment treatment = this.getTreatment(number);
        final ReviewStateTracer state = new ReviewStateTracer(
                        reviewRecords.get(0),
                        true,
                        //we have the toggle event also for OP, but in this case it doesn't toggle, so ignore it
                        treatment == Treatment.OP,
                        treatment == Treatment.TF);
        for (final LogRecord r : reviewRecords) {
            state.handle(r);
        }
        return state.activeTime;
    }

    public long getGrossTimeForReview(int number) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        if (reviewRecords.isEmpty()) {
            return -1;
        }
        final LogRecord firstRecord = reviewRecords.get(0);
        final LogRecord lastRecord = reviewRecords.get(reviewRecords.size() - 1);
        return firstRecord.reviewTimeUntil(lastRecord);
    }

    public long getGrossTimeForUnderstanding(int number) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        final Optional<LogRecord> reviewEnd = reviewRecords.stream().filter((LogRecord r) -> r.getSuffixOfRest(3).startsWith("review ended")).findFirst();
        if (!reviewEnd.isPresent()) {
            return -1;
        }
        final LogRecord last = reviewRecords.get(reviewRecords.size() - 1);
        return reviewEnd.get().reviewTimeUntil(last);
    }

    public List<LogRecord> getReviewRecords(int number) {
        if (this.reviewRecordCache.containsKey(number)) {
            return this.reviewRecordCache.get(number);
        }
        final List<LogRecord> ret = this.records.stream().filter((LogRecord r) -> r.getRest().startsWith("log from review;" + number + ";")).collect(Collectors.toList());
        this.reviewRecordCache.put(number, Collections.unmodifiableList(ret));
        return ret;
    }

    public Treatment getTreatment(int reviewNumber) {
        final String combination = this.getRestAfterMarker("assigned treatment combination ", "NA,NA,NA,NA");
//        System.out.println(combination);
//        System.out.println(shortenTreatment(combination.split(",")[reviewNumber].trim()));
        return shortenTreatment(combination.split(",")[reviewNumber].trim());
    }

    private static Treatment shortenTreatment(String longForm) {
        switch (longForm) {
        case "NA":
            return null;
        case "PROD_FIRST":
            return Treatment.PF;
        case "TEST_FIRST":
            return Treatment.TF;
        case "ONLY_PROD":
            return Treatment.OP;
        default:
            throw new AssertionError("unknown treatment " + longForm);
        }
    }

    public String getTreatmentCombination() {
        final Treatment first = this.getTreatment(0);
        if (first == null) {
            return "NA";
        }
        final Treatment second = this.getTreatment(1);
        return firstIsBetter(first, second) ? (first + "-" + second) : (second + "-" + first);
    }

    public PatchToDiff getChange(int reviewNumber) {
//        if (reviewNumber == 0) {
//            return PatchToDiff.WU;
//        }
        final String combination = this.getRestAfterMarker("assigned treatment combination ", "NA,NA,NA,NA");
//        System.out.println(combination);
//        System.out.println(shortenChange(combination.split(",")[reviewNumber + 2].trim()));
        return shortenChange(combination.split(",")[reviewNumber + 2].trim());
    }

    private static PatchToDiff shortenChange(String longForm) {
        switch (longForm) {
        case "NA":
            return null;
        case "PATCH1":
            return PatchToDiff.P1;
        case "PATCH2":
            return PatchToDiff.P2;
        default:
            throw new AssertionError("unknown change " + longForm);
        }
    }

    public final class RemarkClassifier {

        Map<String, List<Defect>> classification;
        final Map<Integer, String> filenameMap;

        public RemarkClassifier(int reviewNumber) {
            this.classification = SessionData.this.getTrueDefects(SessionData.this.getChange(reviewNumber));
            this.filenameMap = SessionData.this.determineFilenameMap(SessionData.this.getTreatment(reviewNumber));
        }

        public String toRealFilename(String remark) {
            String ret = remark;
            for (final Entry<Integer, String> e : this.filenameMap.entrySet()) {
                ret = ret.replace("#" + e.getKey() + "R;", "#" + e.getValue() + "R;");
                ret = ret.replace("#" + e.getKey() + "L;", "#" + e.getValue() + "L;");
            }
            return ret;
        }

        public List<Defect> getDefects(String remarkWithRealFilename) {
            return this.classification.get(remarkWithRealFilename);
        }

    }

    private Set<Defect> classifyDefects(int reviewNumber) {
        final Collection<String> remarks = this.determineReviewRemarks(reviewNumber);
        final RemarkClassifier classifier = this.createRemarkClassifier(reviewNumber);
        final Set<Defect> defects = new LinkedHashSet<>();
        for (final String remark : remarks) {
            final String remarkWithRealFilename = classifier.toRealFilename(remark);
            final List<Defect> classified = classifier.getDefects(remarkWithRealFilename);
            if (classified == null || classified.isEmpty()) {
                throw new RuntimeException("unclassified remark in " + this.getChange(reviewNumber) + ": " + remarkWithRealFilename);
            } else if (!classified.contains(FP.FP)) {
                defects.addAll(classified);
            } else {
                assert classified.size() == 1;
            }
        }

        return defects;
    }

    public RemarkClassifier createRemarkClassifier(int reviewNumber) {
        return new RemarkClassifier(reviewNumber);
    }

    public Collection<String> getUnclassifiedRemarks() {
        final Set<String> unclassified = new TreeSet<>();
        for (int reviewNumber = 0; reviewNumber < 2; reviewNumber++) {
            final Collection<String> remarks = this.determineReviewRemarks(reviewNumber);
//            System.out.println(remarks);
            if (remarks.isEmpty()) {
                continue;
            }
            final RemarkClassifier classifier = this.createRemarkClassifier(reviewNumber);
            for (final String remark : remarks) {
                final String remarkWithRealFilename = classifier.toRealFilename(remark);
//                System.out.println("Remark is " +remark);
//                System.out.println("Remark with real name is " + remarkWithRealFilename);
                final List<Defect> classified = classifier.getDefects(remarkWithRealFilename);
                if (classified == null || classified.isEmpty()) {
                    unclassified.add("in " + this.getChange(reviewNumber) + ": " + remarkWithRealFilename);
                }
            }
        }
        return unclassified;
    }

    public boolean commentedBothFiles(int reviewNumber){
        if (this.getTreatment(reviewNumber) == Treatment.OP) {
            return true;
        }

        boolean comment0 = false;
        boolean comment1 = false;
        for (final LogRecord r : this.getReviewRecords(reviewNumber)) {
            final String rest = r.getSuffixOfRest(3);
            if (rest.startsWith("startEnterReviewRemark;#0")) {
                comment0 = true;
            }
            if (rest.startsWith("startEnterReviewRemark;#1")) {
                comment1 = true;
            }
        }
        return comment0 && comment1;
    }

    private Map<Integer, String> determineFilenameMap(Treatment treatment) {
        final String rest = this.findFirstStartingWith("hunkMap;" + treatment.getLongForm() + ";c").get().getRest();
        final String map = rest.substring(rest.indexOf('[') + 1, rest.indexOf(']'));
        final Map<Integer, String> ret = new TreeMap<>();
        for (final String part : map.split(",")) {
            final String[] kv = part.trim().split("=");
            ret.put(Integer.parseInt(kv[0]), kv[1]);
        }
        return ret;
    }

    private Collection<String> determineReviewRemarks(int reviewNumber) {
        final Map<String, String> reviewRemarks = new TreeMap<>();
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        if (reviewRecords.isEmpty()) {
            return Collections.emptySet();
        }
        final ReviewStateTracer stateTracer = new ReviewStateTracer(reviewRecords.get(0), true);
        for (final LogRecord r : reviewRecords) {
            stateTracer.handle(r);
            final String rest = r.getSuffixOfRest(3).trim();
            if (rest.startsWith("createReviewRemark;")) {
                reviewRemarks.put(this.getRemarkKey(rest), rest.substring("createReviewRemark;".length()));
            } else if (rest.startsWith("changeReviewRemark;")) {
                reviewRemarks.put(this.getRemarkKey(rest), rest.substring("changeReviewRemark;".length()));
            } else if (rest.startsWith("deleteReviewRemark;")) {
                reviewRemarks.remove(this.getRemarkKey(rest));
            }
        }
        return reviewRemarks.values();
    }

    private String getRemarkKey(String rest) {
        final String[] split = rest.split(";");
        return split[1] + ";" + split[2];
    }

    public String interpretComparisonUnderstanding() {
        if (!(this.isRealEffort(0) && this.isRealEffort(1))) {
            return "NA";
        }
        final String rawValue = this.getVariable("comparisonUnderstanding", null);
        if (rawValue == null) {
            return "NA";
        } else if (rawValue.equals("same")) {
            return rawValue;
        } else if (rawValue.equals("first")) {
            return this.getTreatment(0).toString();
        } else if (rawValue.equals("second")) {
            return this.getTreatment(1).toString();
        } else {
            throw new AssertionError("unexpected answer " + rawValue);
        }
    }

    public String interpretComparisonComplicated() {
        if (!(this.isRealEffort(0) && this.isRealEffort(1))) {
            return "NA";
        }
        final String rawValue = this.getVariable("comparisonComplicated", null);
        if (rawValue == null) {
            return "NA";
        } else if (rawValue.equals("same")) {
            return rawValue;
        } else if (rawValue.equals("first")) {
            return this.getTreatment(0).toString();
        } else if (rawValue.equals("second")) {
            return this.getTreatment(1).toString();
        } else {
            throw new AssertionError("unexpected answer " + rawValue);
        }
    }

    private static boolean firstIsBetter(final Treatment first, final Treatment second) {
        return first.equals(Treatment.TF)
            || (first.equals(Treatment.PF) && second.equals(Treatment.OP));
    }

    public int getWMSpanScore() {
        final List<LogRecord> wmspanRecords = this.getWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        int skips = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    skips++;
                    continue;
                }
                final int lettersCorrect = Integer.parseInt(m.group(1));
                sum += lettersCorrect;
            }
        }
        if (skips != 2) {
            throw new AssertionError("something is wrong with wmspan in " + this.getId() + ", skipped " + skips);
        }
        if (sum <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        return sum;
    }

    public int getWMSpanMathErrors() {
        if (this.getWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.getWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    continue;
                }
                final int mathCorrect = Integer.parseInt(m.group(2));
                sum += size - mathCorrect;
            }
        }
        return sum;
    }

    public long getWMSpanGrossTime() {
        if (this.getWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.getWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        final LogRecord first = wmspanRecords.get(0);
        final LogRecord last = wmspanRecords.get(wmspanRecords.size() - 1);
        return first.wmspanTimeUntil(last);
    }

    private List<LogRecord> getWMSpanRecords() {
        final List<LogRecord> list = this.records.stream().filter((LogRecord r) -> r.getRest().startsWith("log from wm test;")).collect(Collectors.toList());
        //in one case the data is duplicated in the log (most likely, the user pressed back in the browser), discard it
        final List<LogRecord> filtered = new ArrayList<>();
        boolean hadStart = false;
        for (final LogRecord r : list) {
            if (r.getSuffixOfRest(2).trim().equals("operation span initialized")) {
                if (hadStart) {
                    break;
                } else {
                    hadStart = true;
                }
            }
            filtered.add(r);
        }
        assert list.isEmpty() || hadStart;
        return filtered;
    }

    /**
     * Checks if there were any revisits of an already opened page and returns these
     * states/pages. Consequently, an empty list is returned if everything is OK.
     */
    public List<String> determinePageReloads() {
        final Set<String> visited = new HashSet<>();
        final List<String> ret = new ArrayList<>();
        for (final LogRecord record : this.records) {
            if (record.getRest().startsWith("change state;")) {
                final String state = record.getSuffixOfRest(1);
                if (visited.contains(state)) {
                    ret.add(state);
                } else {
                    visited.add(state);
                }
            }
        }
        return ret;
    }

    private Map<String, List<Defect>> getTrueDefects(PatchToDiff changeId) {
        switch (changeId) {
        case P1:
            return P1.getTrueDefects();
        case P2:
            return P2.getTrueDefects();
        default:
            throw new AssertionError("unknown id " + changeId);
        }
    }

    public String getLastState() {
        final ListIterator<LogRecord> iter = this.records.listIterator(this.records.size());
        while (iter.hasPrevious()) {
            final LogRecord r = iter.previous();
            if (r.getRest().startsWith("change state;")) {
                return r.getRestField(1);
            }
        }
        throw new AssertionError();
    }

    /**
     * Return true if it seems the participant did a real effort on the review with the given number.
     * It is not a real effort if there are neither remarks nor correctly answered understanding questions
     * and the duration is very short.
     */
    public boolean isRealEffort(int reviewNumber) {
        final long duration = this.getNetTimeForReview(reviewNumber, true);
        if (duration >= 5L * 60 * 1000) {
            return true;
        }
        final Collection<String> remarks = this.determineReviewRemarks(reviewNumber);
        if (!remarks.isEmpty()) {
            return true;
        }
        return false;
    }

    public int foundDefect(int reviewNumber, Defect d) {
        if (!this.isRealEffort(reviewNumber)) {
            return -1;
        }
        return this.classifyDefects(reviewNumber).contains(d) ? 1 : 0;
    }

    public int getReviewNumberWith(PatchToDiff diff) {
        if (this.getChange(0) == diff) {
            return 0;
        } else //            assert !this.isRealEffort(1);
            //            return 1;
            if (this.getChange(1) == diff) {
            return 1;
        } else {
                throw new java.lang.RuntimeException();
            }
    }

    public int getViewWidth() {
        final Pattern p = Pattern.compile("window size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(1));
    }

    private Matcher findMatchInReviewData(int reviewNumber, Pattern pattern) {
        final List<LogRecord> review = this.getReviewRecords(reviewNumber);
        for (final LogRecord r : review) {
            final String suffix = r.getSuffixOfRest(3).trim();
            final Matcher m = pattern.matcher(suffix);
            if (m.matches()) {
                return m;
            }
        }
        return null;
    }

    public int getViewHeight() {
        return this.getViewHeight(1);
    }

    public int getViewHeight(int reviewNumber) {
        final Pattern p = Pattern.compile("window size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(reviewNumber, p);
        return m == null ? -1 : Integer.parseInt(m.group(2));
    }

    public int getScreenWidth() {
        final Pattern p = Pattern.compile("screen size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(1));
    }

    public int getScreenHeight() {
        final Pattern p = Pattern.compile("screen size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(2));
    }

    public String getBrowser() {
        final Pattern p = Pattern.compile("view loaded,user agent=(.+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? "NA" : this.translateBrowser(m.group(1));
    }

    private String translateBrowser(String userAgent) {
        if (userAgent.contains("OPR/")) {
            return "Opera";
        } else if (userAgent.contains("Trident/")) {
            return "IE";
        } else if (userAgent.contains("Chrome/")) {
            return "Chrome";
        } else if (userAgent.contains("Safari/")) {
            return "Safari";
        } else if (userAgent.contains("Firefox/")) {
            return "Firefox";
        }
        throw new AssertionError("unknown user agent " + userAgent);
    }

    public String getExternalIpAtStart() {
        final String record = this.getRestAfterMarker("request for ", null);
        final int index = record.indexOf(" from ");
        if (index < 0) {
            return "NA";
        }
        return record.substring(index + 6).trim();
    }

    public String getInternalIp() {
        return this.getVariable("clientSideIpHash", "NA");
    }

    public boolean isPrize() {
        return Long.parseLong(this.getId()) > 238915110303L;
    }

    public boolean isControlledSetting() {
        final String extIp = this.getExternalIpAtStart();
        return extIp.equals("1A6CB90A4B8DBACECD2DD4F5BEBE6095")
            || (extIp.equals("NA") && !this.isPrize() && this.getBrowser().equals("IE"))
            || this.getId().equals("237967248580");
    }

    public String getPreviousSessionR1FromCookie() {
        final Pattern p = Pattern.compile("cookie;lastReview0=[0-9]+; lastReview1=([0-9]+).*");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? "NA" : m.group(1);
    }

    public boolean isDuplicateParticipation() {
        return this.DUPLICATE_BLACKLIST.contains(this.getId());
    }

    public List<String> getExceptions() {
        return this.records.stream()
                        .filter((LogRecord l) -> l.getRest().startsWith("exception;"))
                        .map((LogRecord l) -> l.getSuffixOfRest(1))
                        .collect(Collectors.toList());
    }

    public long getScrollAmount(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        long currentY = 0;
        long scrollSum = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3);
            if (key.equals("scroll")) {
                final String y = r.getRestField(5).trim();
                final long nextY = Math.round(Double.parseDouble(y));
                scrollSum += Math.abs(currentY - nextY);
                currentY = nextY;
            }
        }
        return scrollSum;
    }

    public int getProdMaintainabilityIssues(int reviewNumber){
        int ret = 0;
        ret += this.classifyDefects(reviewNumber).contains(P1.PROD_METHOD_DESCRIPTION) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P1.INITIALIZATION_WRONG_PLACE) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P1.METHOD_SIGNATURE_CHANGE) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.VARIABLE_NAMES) ? 1 : 0;
        return ret;
    }

    public int getTestMaintainabilityIssues(int reviewNumber){
        int ret = 0;
        ret += this.classifyDefects(reviewNumber).contains(P1.SQUARE_GLOBAL_VAR) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.TEST_METHOD_DESCRIPTION) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.INIT_SPRITE) ? 1 : 0;
        return ret;
    }

    public int getProdBugs(int reviewNumber){
        int ret = 0;
        ret += this.classifyDefects(reviewNumber).contains(P1.NULL_PLAYER) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.X_Y_GREATER_THAN_ZERO) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.WRONG_RETURN_TYPE) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.TYPO) ? 1 : 0;
        return ret;
    }

    public int getTestBugs(int reviewNumber){
        int ret = 0;
        ret += this.classifyDefects(reviewNumber).contains(P1.WRONG_SQUARE2) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.DIFFERENT_NUMBERS) ? 1 : 0;
        ret += this.classifyDefects(reviewNumber).contains(P2.WRONG_ASSERT) ? 1 : 0;
        return ret;
    }

    public int getTotalNumberOfIssues(int reviewNumber, String issue){
        switch (issue){
        case "ProdMaint":
            return this.getChange(reviewNumber).getProdMaint();
        case "TestMaint":
            return this.getTreatment(reviewNumber) == Treatment.OP ? 0 : this.getChange(reviewNumber).getTestMaint();
        case "ProdBugs":
            return this.getChange(reviewNumber).getProdBugs();
        case "TestBugs":
            return this.getTreatment(reviewNumber) == Treatment.OP ? 0 : this.getChange(reviewNumber).getTestBugs();
        default:
            throw new RuntimeException();
        }
    }

    public long getPauseCount(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        int ret = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3).trim();
            if (key.equals("pause started")) {
                ret++;
            }
        }
        return ret;
    }

    public long getDescriptionCount(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        int ret = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3).trim();
            if (key.equals("desc started")) {
                ret++;
            }
        }
        return ret;
    }

    public long getToggleCount(int reviewNumber) {
        if (this.getTreatment(reviewNumber) == Treatment.OP) {
            return -1;
        }
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        int ret = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3).trim();
            if (key.equals("pressedToggleShownCode")) {
                ret++;
            }
        }
        return ret;
    }

}
